// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package commands

import (
	"errors"
	"github.com/jessevdk/go-flags"
	"gitlab.schukai.com/oss/utilities/documentation-manager/environment"
	"gitlab.schukai.com/oss/utilities/documentation-manager/translations"
)

const (
	// Generic codes.
	CodeOK      = 0 // success
	CodeErr     = 1 // generic error
	CodeHelpErr = 2 // command is invoked with -help or -h flag but no such flag is defined

	// Codes as defined in /usr/include/sysexits.h on *nix systems.
	CodeUsage       = 64 // command line usage error
	CodeDataErr     = 65 // data format error
	CodeNoInput     = 66 // cannot open input
	CodeNoUser      = 67 // addressee unknown
	CodeNoHost      = 68 // host name unknown
	CodeUnavailable = 69 // service unavailable
	CodeSoftware    = 70 // internal software error
	CodeOSErr       = 71 // system error (e.g., can't fork)
	CodeOSFile      = 72 // critical OS file missing
	CodeCantCreat   = 73 // can't create (user) output file
	CodeIOErr       = 74 // input/output error
	CodeTempFail    = 75 // temp failure; user is invited to retry
	CodeProtocol    = 76 // remote error in protocol
	CodeNoPerm      = 77 // permission denied
	CodeConfig      = 78 // configuration error

)

var (
	notPermittedError = errors.New(translations.T.Sprintf("Not permitted"))
)

func checkFlagsError(err error) {
	e := err.(*flags.Error)
	switch e.Type {
	case flags.ErrUnknown:
		environment.State.AddError(translations.T.Sprintf("flag error unknown")).Exit()
	case flags.ErrExpectedArgument:
		environment.State.AddError(translations.T.Sprintf("expected argument")).Exit()
	case flags.ErrUnknownFlag:
		environment.State.AddError(translations.T.Sprintf("unknown flag")).Exit()
	case flags.ErrUnknownGroup:
		environment.State.AddError(translations.T.Sprintf("unknown group")).Exit()
	case flags.ErrMarshal:
		environment.State.AddError(translations.T.Sprintf("marshal")).Exit()
	case flags.ErrHelp:
		environment.State.Exit()
	case flags.ErrNoArgumentForBool:
		environment.State.AddError(translations.T.Sprintf("no argument for bool")).Exit()
	case flags.ErrRequired:
		environment.State.Exit()
	case flags.ErrShortNameTooLong:
		environment.State.AddError(translations.T.Sprintf("short name too long")).Exit()
	case flags.ErrDuplicatedFlag:
		environment.State.AddError(translations.T.Sprintf("duplicated flag")).Exit()
	case flags.ErrTag:
		environment.State.AddError(translations.T.Sprintf("tag %s", err.Error())).Exit()
	case flags.ErrCommandRequired:
		environment.State.Exit()
	case flags.ErrUnknownCommand:
		environment.State.AddError(translations.T.Sprintf("unknown command")).Exit()
	case flags.ErrInvalidChoice:
		environment.State.AddError(translations.T.Sprintf("invalid choice")).Exit()
	case flags.ErrInvalidTag:
		environment.State.AddError(translations.T.Sprintf("invalid tag")).Exit()
	default:
		environment.State.AddError(translations.T.Sprintf("unrecognized error type")).Exit()

	}
}

func checkError(err error) {
	if err == nil {
		return
	}

	switch err.(type) {
	case *flags.Error:
		checkFlagsError(err)
	}

	environment.ExitWithError(1, "Unknown Error: %s", err.Error())
}
